@@ -12,12 +12,11 @@
 QT_BEGIN_NAMESPACE
 
 #ifndef QT_NO_DEBUG
-static const bool qsg_leak_check = !qEnvironmentVariableIsEmpty("QML_LEAK_CHECK");
 static int qt_node_count = 0;
 
 static void qt_print_node_count()
 {
-    qDebug("Number of leaked nodes: %i", qt_node_count);
+    qCDebug(lcQsgLeak, "Number of leaked nodes: %i", qt_node_count);
     qt_node_count = -1;
 }
 #endif
@@ -42,8 +41,8 @@ static void qt_print_node_count()
     insertChildNodeAfter(). The order of nodes is important as geometry nodes
     are rendered according to their ordering in the scene graph.
 
-    The scene graph nodes contains a mechanism to describe which
-    parts of the scene has changed. This includes the combined matrices,
+    The scene graph nodes contain a mechanism that describes which
+    parts of the scene have changed. This includes the combined matrices,
     accumulated opacity, changes to the node hierarchy, and so on. This
     information can be used for optimizations inside the scene graph renderer.
     For the renderer to properly render the nodes, it is important that users
@@ -256,7 +255,7 @@ QSGNode::QSGNode(QSGNodePrivate &dd, NodeType type)
 void QSGNode::init()
 {
 #ifndef QT_NO_DEBUG
-    if (qsg_leak_check) {
+    if (lcQsgLeak().isDebugEnabled()) {
         ++qt_node_count;
         static bool atexit_registered = false;
         if (!atexit_registered) {
@@ -281,10 +280,10 @@ void QSGNode::init()
 QSGNode::~QSGNode()
 {
 #ifndef QT_NO_DEBUG
-    if (qsg_leak_check) {
+    if (lcQsgLeak().isDebugEnabled()) {
         --qt_node_count;
         if (qt_node_count < 0)
-            qDebug("Node destroyed after qt_print_node_count() was called.");
+            qCDebug(lcQsgLeak, "Node destroyed after qt_print_node_count() was called.");
     }
 #endif
     destroy();
