@@ -7,7 +7,7 @@
 #include <QtCore/QWaitCondition>
 #include <QtCore/QAnimationDriver>
 #include <QtCore/QQueue>
-#include <QtCore/QTime>
+#include <QtCore/QTimer>
 
 #include <QtGui/QGuiApplication>
 #include <QtGui/QScreen>
@@ -18,6 +18,7 @@
 #include <QtQuick/QQuickWindow>
 #include <private/qquickwindow_p.h>
 #include <private/qquickitem_p.h>
+#include <QtGui/qpa/qplatformwindow_p.h>
 
 #include <QtQuick/private/qsgrenderer_p.h>
 
@@ -234,8 +235,8 @@ public:
         , stopEventProcessing(false)
     {
         sgrc = static_cast<QSGDefaultRenderContext *>(renderContext);
-#if (defined(Q_OS_QNX) && defined(Q_PROCESSOR_X86)) || defined(Q_OS_INTEGRITY)
-        // The SDP 6.6.0 x86 MESA driver requires a larger stack than the default.
+#if defined(Q_OS_QNX) || defined(Q_OS_INTEGRITY)
+        // The render thread requires a larger stack than the default (256k).
         setStackSize(1024 * 1024);
 #endif
     }
@@ -1461,10 +1462,26 @@ void QSGThreadedRenderLoop::update(QQuickWindow *window)
     if (!w)
         return;
 
-    if (w->thread == QThread::currentThread()) {
-        qCDebug(QSG_LOG_RENDERLOOP) << "update on window - on render thread" << w->window;
-        w->thread->requestRepaint();
-        return;
+    const bool isRenderThread = QThread::currentThread() == w->thread;
+
+    if (QPlatformWindow *platformWindow = window->handle()) {
+        // If the window is being resized we don't want to schedule unthrottled
+        // updates on the render thread, as this will starve the main thread
+        // from getting drawables for displaying the updated window size.
+        if (isRenderThread && !platformWindow->allowsIndependentThreadedRendering()) {
+            // In most cases the window will already have update requested
+            // due to the animator triggering a sync, but just in case we
+            // schedule an update request on the main thread explicitly.
+            qCDebug(QSG_LOG_RENDERLOOP) << "window is resizing. update on window" << w->window;
+            QTimer::singleShot(0, window, [=]{ window->requestUpdate(); });
+            return;
+        }
+    }
+
+    if (isRenderThread) {
+       qCDebug(QSG_LOG_RENDERLOOP) << "update on window - on render thread" << w->window;
+       w->thread->requestRepaint();
+       return;
     }
 
     qCDebug(QSG_LOG_RENDERLOOP) << "update on window" << w->window;
@@ -1664,25 +1681,47 @@ void QSGThreadedRenderLoop::polishAndSync(Window *w, bool inExpose)
     // isVSyncDependent() == true, if not then we always use the driver and
     // just advance here)
     if (m_animation_timer == 0 && m_animation_driver->isRunning()) {
-        qCDebug(QSG_LOG_RENDERLOOP, "- advancing animations");
-        m_animation_driver->advance();
-        qCDebug(QSG_LOG_RENDERLOOP, "- animations done..");
+        auto advanceAnimations = [this, window=QPointer(window)] {
+            qCDebug(QSG_LOG_RENDERLOOP, "- advancing animations");
+            m_animation_driver->advance();
+            qCDebug(QSG_LOG_RENDERLOOP, "- animations done..");
 
-        // We need to trigger another update round to keep all animations
-        // running correctly. For animations that lead to a visual change (a
-        // property change in some item leading to dirtying the item and so
-        // ending up in maybeUpdate()) this would not be needed, but other
-        // animations would then stop functioning since there is nothing
-        // advancing the animation system if we do not call postUpdateRequest()
-        // here and nothing else leads to it either. This has an unfortunate
-        // side effect in multi window cases: one can end up in a situation
-        // where a non-animating window gets updates continuously because there
-        // is an animation running in some other window that is non-exposed or
-        // even closed already (if it was exposed we would not hit this branch,
-        // however). Sadly, there is nothing that can be done about it.
-        postUpdateRequest(w);
+            // We need to trigger another update round to keep all animations
+            // running correctly. For animations that lead to a visual change (a
+            // property change in some item leading to dirtying the item and so
+            // ending up in maybeUpdate()) this would not be needed, but other
+            // animations would then stop functioning since there is nothing
+            // advancing the animation system if we do not call postUpdateRequest()
+            // here and nothing else leads to it either. This has an unfortunate
+            // side effect in multi window cases: one can end up in a situation
+            // where a non-animating window gets updates continuously because there
+            // is an animation running in some other window that is non-exposed or
+            // even closed already (if it was exposed we would not hit this branch,
+            // however). Sadly, there is nothing that can be done about it.
+            if (window)
+                window->requestUpdate();
 
-        emit timeToIncubate();
+            emit timeToIncubate();
+        };
+
+#if defined(Q_OS_APPLE)
+        if (inExpose) {
+            // If we are handling an expose event the system is expecting us to
+            // produce a frame that it can present on screen to the user. Advancing
+            // animations at this point might result in changing properties of the
+            // window in a way that invalidates the current frame, resulting in the
+            // discarding of the current frame before the user ever sees it. To give
+            // the system a chance to present the current frame we defer the advance
+            // of the animations until the start of the next event loop pass, which
+            // should still give plenty of time to compute the new render state before
+            // the next expose event or update request.
+            QMetaObject::invokeMethod(this, advanceAnimations, Qt::QueuedConnection);
+        } else
+#endif // Q_OS_APPLE
+        {
+            // For regular update requests we assume we can advance here synchronously
+            advanceAnimations();
+        }
     } else if (w->updateDuringSync) {
         postUpdateRequest(w);
     }
@@ -1714,6 +1753,7 @@ bool QSGThreadedRenderLoop::event(QEvent *e)
             emit timeToIncubate();
             return true;
         }
+        break;
     }
 
     default:
