@@ -56,7 +56,7 @@
 
 QT_BEGIN_NAMESPACE
 
-Q_DECLARE_LOGGING_CATEGORY(lcQsgLeak)
+using namespace Qt::Literals::StringLiterals;
 
 #if defined(QT_DEBUG) && QT_CONFIG(thread)
 class ThreadAffinityMarker
@@ -96,7 +96,7 @@ private:
 
 const QLatin1String QQuickPixmap::itemGrabberScheme = QLatin1String("itemgrabber");
 
-Q_LOGGING_CATEGORY(lcImg, "qt.quick.image")
+Q_STATIC_LOGGING_CATEGORY(lcImg, "qt.quick.image")
 
 /*! \internal
     The maximum currently-unused image data that can be stored for potential
@@ -303,7 +303,7 @@ class QQuickPixmapData
 public:
     QQuickPixmapData(const QUrl &u, const QRect &r, const QSize &rs,
                      const QQuickImageProviderOptions &po, const QString &e)
-    : refCount(1), frameCount(1), frame(0), inCache(false), pixmapStatus(QQuickPixmap::Error),
+    : refCount(1), frameCount(1), frame(0), inCache(false), fromSpecialDevice(false), pixmapStatus(QQuickPixmap::Error),
       url(u), errorString(e), requestRegion(r), requestSize(rs),
       providerOptions(po), appliedTransform(QQuickImageProviderOptions::UsePluginDefaultTransform),
       textureFactory(nullptr), reply(nullptr), prevUnreferenced(nullptr),
@@ -316,7 +316,7 @@ public:
 
     QQuickPixmapData(const QUrl &u, const QRect &r, const QSize &s, const QQuickImageProviderOptions &po,
                      QQuickImageProviderOptions::AutoTransform aTransform, int frame=0, int frameCount=1)
-    : refCount(1), frameCount(frameCount), frame(frame), inCache(false), pixmapStatus(QQuickPixmap::Loading),
+    : refCount(1), frameCount(frameCount), frame(frame), inCache(false), fromSpecialDevice(false), pixmapStatus(QQuickPixmap::Loading),
       url(u), requestRegion(r), requestSize(s),
       providerOptions(po), appliedTransform(aTransform),
       textureFactory(nullptr), reply(nullptr), prevUnreferenced(nullptr), prevUnreferencedPtr(nullptr),
@@ -330,7 +330,7 @@ public:
     QQuickPixmapData(const QUrl &u, QQuickTextureFactory *texture,
                      const QSize &s, const QRect &r, const QSize &rs, const QQuickImageProviderOptions &po,
                      QQuickImageProviderOptions::AutoTransform aTransform, int frame=0, int frameCount=1)
-    : refCount(1), frameCount(frameCount), frame(frame), inCache(false), pixmapStatus(QQuickPixmap::Ready),
+    : refCount(1), frameCount(frameCount), frame(frame), inCache(false), fromSpecialDevice(false), pixmapStatus(QQuickPixmap::Ready),
       url(u), implicitSize(s), requestRegion(r), requestSize(rs),
       providerOptions(po), appliedTransform(aTransform),
       textureFactory(texture), reply(nullptr), prevUnreferenced(nullptr),
@@ -342,7 +342,7 @@ public:
     }
 
     QQuickPixmapData(QQuickTextureFactory *texture)
-    : refCount(1), frameCount(1), frame(0), inCache(false), pixmapStatus(QQuickPixmap::Ready),
+    : refCount(1), frameCount(1), frame(0), inCache(false), fromSpecialDevice(false), pixmapStatus(QQuickPixmap::Ready),
       appliedTransform(QQuickImageProviderOptions::UsePluginDefaultTransform),
       textureFactory(texture), reply(nullptr), prevUnreferenced(nullptr),
       prevUnreferencedPtr(nullptr), nextUnreferenced(nullptr)
@@ -370,6 +370,7 @@ public:
     int frame;
 
     bool inCache:1;
+    bool fromSpecialDevice:1;
 
     QQuickPixmap::Status pixmapStatus;
     QUrl url;
@@ -381,7 +382,7 @@ public:
     QQuickImageProviderOptions::AutoTransform appliedTransform;
     QColorSpace targetColorSpace;
 
-    QIODevice *specialDevice = nullptr;
+    QPointer<QIODevice> specialDevice;
 
     // actual image data, after loading
     QQuickTextureFactory *textureFactory;
@@ -606,20 +607,18 @@ QQuickPixmapReader::~QQuickPixmapReader()
             delete reply;
         }
         jobs.clear();
-#if QT_CONFIG(qml_network)
         const auto cancelJob = [this](QQuickPixmapReply *reply) {
             if (reply->loading) {
                 cancelledJobs.append(reply);
                 reply->data = nullptr;
             }
         };
-
+#if QT_CONFIG(qml_network)
         for (auto *reply : std::as_const(networkJobs))
             cancelJob(reply);
-
+#endif
         for (auto *reply : std::as_const(asyncResponses))
             cancelJob(reply);
-#endif
 #if !QT_CONFIG(quick_pixmap_cache_threaded_download)
     // In this case we won't be waiting, but we are on the correct thread already, so we can
     // perform housekeeping synchronously now.
@@ -640,7 +639,6 @@ QQuickPixmapReader::~QQuickPixmapReader()
     wait();
 #endif
 
-#if QT_CONFIG(qml_network)
     // While we've been waiting, the other thread may have added
     // more replies. No one will care about them anymore.
 
@@ -649,16 +647,17 @@ QQuickPixmapReader::~QQuickPixmapReader()
             reply->data->reply = nullptr;
         delete reply;
     };
-
+#if QT_CONFIG(qml_network)
     for (QQuickPixmapReply *reply : std::as_const(networkJobs))
         deleteReply(reply);
-
+#endif
     for (QQuickPixmapReply *reply : std::as_const(asyncResponses))
         deleteReply(reply);
 
+#if QT_CONFIG(qml_network)
     networkJobs.clear();
-    asyncResponses.clear();
 #endif
+    asyncResponses.clear();
 }
 
 #if QT_CONFIG(qml_network)
@@ -803,7 +802,9 @@ void QQuickPixmapReader::processJobs()
                         // cancel any jobs already started
                         reply->close();
                     }
-                } else {
+                } else
+#endif
+                {
                     QQuickImageResponse *asyncResponse = asyncResponses.key(job);
                     if (asyncResponse) {
                         asyncResponses.remove(asyncResponse);
@@ -811,7 +812,6 @@ void QQuickPixmapReader::processJobs()
                     }
                 }
                 PIXMAP_PROFILE(pixmapStateChanged<QQuickProfiler::PixmapLoadingError>(job->url));
-#endif
                 // deleteLater, since not owned by this thread
                 job->deleteLater();
             }
@@ -1006,9 +1006,22 @@ void QQuickPixmapReader::processJob(QQuickPixmapReply *runningJob, const QUrl &u
             QString errorStr;
             QSize readSize;
 
-            if (runningJob->data && runningJob->data->specialDevice) {
+            if (runningJob->data && runningJob->data->fromSpecialDevice) {
+                auto specialDevice = runningJob->data->specialDevice;
+                if (specialDevice.isNull() || QObjectPrivate::get(specialDevice.data())->deleteLaterCalled) {
+                    qCDebug(lcImg) << "readImage job aborted" << url;
+                    return;
+                }
                 int frameCount;
-                if (!readImage(url, runningJob->data->specialDevice, &image, &errorStr, &readSize, &frameCount,
+                // Ensure that specialDevice's thread affinity is _this_ thread, to avoid deleteLater()
+                // deleting prematurely, before readImage() is done. But this is only possible if it has already
+                // relinquished its initial thread affinity.
+                if (!specialDevice->thread()) {
+                    qCDebug(lcQsgLeak) << specialDevice.data() << ": changing thread affinity so that"
+                                       << QThread::currentThread() << "will handle any deleteLater() calls";
+                    specialDevice->moveToThread(QThread::currentThread());
+                }
+                if (!readImage(url, specialDevice.data(), &image, &errorStr, &readSize, &frameCount,
                                runningJob->requestRegion, runningJob->requestSize,
                                runningJob->providerOptions, nullptr, runningJob->data->frame)) {
                     errorCode = QQuickPixmapReply::Loading;
@@ -1495,7 +1508,7 @@ void QQuickPixmapData::addToCache()
             qCDebug(lcImg) << "adding" << key << "to total" << QQuickPixmapCache::instance()->m_cache.size();
             for (auto it = QQuickPixmapCache::instance()->m_cache.keyBegin(); it != QQuickPixmapCache::instance()->m_cache.keyEnd(); ++it) {
                 if (*(it->url) == url && it->frame == frame)
-                    qDebug(lcImg) << "    similar pre-existing:" << *it;
+                    qCDebug(lcImg) << "    similar pre-existing:" << *it;
             }
         }
         QQuickPixmapCache::instance()->m_cache.insert(key, this);
@@ -1959,6 +1972,7 @@ void QQuickPixmap::loadImageFromDevice(QQmlEngine *engine, QIODevice *device, co
         d = new QQuickPixmapData(url, requestRegion, requestSize, providerOptions,
                                  QQuickImageProviderOptions::UsePluginDefaultTransform, frame, frameCount);
         d->specialDevice = device;
+        d->fromSpecialDevice = true;
         d->addToCache();
 
         QQuickPixmapReader::readerMutex.lock();
@@ -2008,6 +2022,17 @@ bool QQuickPixmap::isCached(const QUrl &url, const QRect &requestRegion, const Q
     return store->m_cache.contains(key);
 }
 
+bool QQuickPixmap::isScalableImageFormat(const QUrl &url)
+{
+    if (url.scheme() == "image"_L1)
+        return true;
+
+    const QString stringUrl = url.path(QUrl::PrettyDecoded);
+    return stringUrl.endsWith("svg"_L1)
+            || stringUrl.endsWith("svgz"_L1)
+            || stringUrl.endsWith("pdf"_L1);
+}
+
 bool QQuickPixmap::connectFinished(QObject *object, const char *method)
 {
     if (!d || !d->reply) {
@@ -2061,4 +2086,5 @@ QT_END_NAMESPACE
 
 #include <qquickpixmapcache.moc>
 
+#include "moc_qquickpixmap_p.cpp"
 #include "moc_qquickpixmapcache_p.cpp"
