@@ -1,5 +1,5 @@
 // Copyright (C) 2022 The Qt Company Ltd.
-// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
 
 #include <QtTest/QtTest>
 #include <QtCore/QCoreApplication>
@@ -33,6 +33,8 @@ private slots:
     void siblings();
     void initial();
     void noApplicationIdentifiersSet();
+    void fromResources();
+    void overridenSettings();
 };
 
 // ### Replace keyValueMap("foo", "bar") with QVariantMap({{"foo", "bar"}})
@@ -493,6 +495,38 @@ void tst_QQmlSettings::noApplicationIdentifiersSet()
     QVERIFY(!settings.value("success").toBool());
 }
 
+void tst_QQmlSettings::fromResources()
+{
+    QQmlEngine engine;
+    QQmlComponent component(&engine, testFileUrl("resources.qml"));
+    QVERIFY2(component.isReady(), qPrintable(component.errorString()));
+    QScopedPointer<QObject> root(component.create());
+    QVERIFY(!root.isNull());
+
+    QCOMPARE(root->property("text").toString(), QLatin1String("from resource"));
+}
+
+void tst_QQmlSettings::overridenSettings() {
+    QQmlEngine engine;
+    QQmlComponent component(&engine, testFileUrl("override.qml"));
+    QVERIFY2(component.isReady(), qPrintable(component.errorString()));
+    delete component.create();
+
+    QSettings qs;
+
+    QVERIFY(qs.childGroups().contains("Parent"));
+    qs.beginGroup("Parent");
+    QVERIFY(qs.childKeys().contains("overriddenProperty"));
+    QCOMPARE(qs.value("overriddenProperty").toString(), QStringLiteral("parentChanged"));
+    qs.endGroup();
+
+    QVERIFY(qs.childGroups().contains("Child"));
+    qs.beginGroup("Child");
+    QVERIFY(qs.childKeys().contains("overriddenProperty"));
+    QCOMPARE(qs.value("overriddenProperty").toString(), QStringLiteral("childChanged"));
+    qs.endGroup();
+}
+
 QTEST_MAIN(tst_QQmlSettings)
 
 #include "tst_qqmlsettings.moc"
