@@ -3,17 +3,23 @@
 
 /*---
 esid: sec-array.prototype.some
-es5id: 15.4.4.17-8-5
+es5id: 15.4.4.17-8-6
 description: >
     Array.prototype.some returns false if 'length' is 0 (subclassed
-    Array, length overridden to '0' (type conversion))
+    Array, length overridden with obj with valueOf)
 ---*/
 
 foo.prototype = new Array(1, 2, 3);
 
 function foo() {}
 var f = new foo();
-f.length = '0';
+
+var o = {
+  valueOf: function() {
+    return 0;
+  }
+};
+f.length = o;
 
 function cb() {}
 var i = f.some(cb);
