@@ -4,14 +4,17 @@
 #include "WithBindableProperties.h"
 
 #include <private/qmlutils_p.h>
+#include <private/qqmlanybinding_p.h>
 #include <private/qqmlbind_p.h>
 #include <private/qqmlcomponentattached_p.h>
+#include <private/qqmlpropertytopropertybinding_p.h>
 #include <private/qquickrectangle_p.h>
 
 #include <QtTest/qtest.h>
 
-#include <QtQml/qqmlengine.h>
 #include <QtQml/qqmlcomponent.h>
+#include <QtQml/qqmlengine.h>
+#include <QtQml/qqmlproperty.h>
 
 class tst_qqmlbinding : public QQmlDataTest
 {
@@ -46,6 +49,7 @@ private slots:
     void whenEvaluatedEarlyEnough();
     void propertiesAttachedToBindingItself();
     void toggleEnableProperlyRemembersValues();
+    void qQmlPropertyToPropertyBinding();
 
 private:
     QQmlEngine engine;
@@ -674,6 +678,117 @@ void tst_qqmlbinding::toggleEnableProperlyRemembersValues()
     }
 }
 
+class PropertyToPropertyObject : public QObject
+{
+    Q_OBJECT
+    Q_PROPERTY(QRectF a READ a WRITE setA NOTIFY aChanged)
+    Q_PROPERTY(QRectF b READ b BINDABLE bindableB)
+
+    Q_PROPERTY(qreal top READ top WRITE setTop NOTIFY topChanged)
+    Q_PROPERTY(qreal bottom READ bottom BINDABLE bindableBottom)
+    Q_PROPERTY(qreal left READ left WRITE setLeft NOTIFY leftChanged)
+    Q_PROPERTY(qreal right READ right BINDABLE bindableRight)
+public:
+
+    QRectF a() const { return m_a; }
+    void setA(const QRectF &a)
+    {
+        if (m_a == a)
+            return;
+        m_a = a;
+        emit aChanged();
+    }
+
+    QRectF b() const { return m_b.value(); }
+    QBindable<QRectF> bindableB() { return QBindable<QRectF>(&m_b); }
+
+    qreal top() const { return m_top; }
+    void setTop(qreal top)
+    {
+        if (qFuzzyCompare(m_top, top))
+            return;
+        m_top = top;
+        emit topChanged();
+    }
+
+    qreal bottom() const { return m_bottom; }
+    QBindable<qreal> bindableBottom() { return QBindable<qreal>(&m_bottom); }
+
+    qreal left() const { return m_left; }
+    void setLeft(qreal left)
+    {
+        if (qFuzzyCompare(m_left, left))
+            return;
+        m_left = left;
+        emit leftChanged();
+    }
+
+    qreal right() const { return m_right; }
+    QBindable<qreal> bindableRight() { return QBindable<qreal>(&m_right); }
+
+signals:
+    void aChanged();
+    void topChanged();
+    void leftChanged();
+
+private:
+    QRectF m_a;
+    QProperty<QRectF> m_b;
+    qreal m_top = 0;
+    QProperty<qreal> m_bottom;
+    qreal m_left = 0;
+    QProperty<qreal> m_right;
+};
+
+void tst_qqmlbinding::qQmlPropertyToPropertyBinding()
+{
+    using namespace Qt::StringLiterals;
+
+    QScopedPointer<PropertyToPropertyObject> target(new PropertyToPropertyObject);
+    QScopedPointer<PropertyToPropertyObject> source(new PropertyToPropertyObject);
+
+    auto installPropertyBinding
+            = [&](QObject *targetObject, const QString &targetPropertyName,
+                  QObject *sourceObject, const QString &sourcePropertyName) {
+        QQmlProperty targetProperty(targetObject, targetPropertyName);
+        QQmlProperty sourceProperty(sourceObject, sourcePropertyName);
+        QQmlAnyBinding binding;
+
+        binding = new QQmlPropertyToPropertyBinding(
+                &engine,
+                sourceObject, QQmlPropertyPrivate::get(sourceProperty)->encodedIndex(),
+                targetObject, targetProperty.index());
+        binding.installOn(targetProperty);
+    };
+
+    installPropertyBinding(target.data(), "left"_L1,   source.data(), "a.left"_L1);
+    installPropertyBinding(target.data(), "top"_L1,    source.data(), "b.top"_L1);
+
+    // TODO: We cannot install a QQmlPropertyToPropertyBinding on bindable properties
+    //       This is because it derives from QQmlAbstractBinding which is only meant for
+    //       non-bindable properties.
+
+    // installPropertyBinding(target.data(), "right"_L1,  source.data(), "a.right"_L1);
+    // installPropertyBinding(target.data(), "bottom"_L1, source.data(), "b.bottom"_L1);
+
+    QCOMPARE(target->top(), 0);
+    QCOMPARE(target->bottom(), 0);
+    QCOMPARE(target->left(), 0);
+    QCOMPARE(target->right(), 0);
+
+    source->setA(QRectF(11, 22, 33, 44));
+    QCOMPARE(target->top(), 0);
+    QCOMPARE(target->bottom(), 0);
+    QCOMPARE(target->left(), 11);
+    // QCOMPARE(target->right(), 33);
+
+    source->bindableB().setValue(QRectF(55, 66, 77, 88));
+    QCOMPARE(target->top(), 66);
+    // QCOMPARE(target->bottom(), 88);
+    QCOMPARE(target->left(), 11);
+    // QCOMPARE(target->right(), 33);
+}
+
 QTEST_MAIN(tst_qqmlbinding)
 
 #include "tst_qqmlbinding.moc"
