@@ -133,6 +133,8 @@ private slots:
     void boundComponent();
     void loadFromModule_data();
     void loadFromModule();
+    void loadFromModuleSyncAndAsync_data();
+    void loadFromModuleSyncAndAsync();
     void loadFromModuleLifecycle();
     void loadFromModuleThenCreateWithIncubator();
     void loadFromModuleFailures_data();
@@ -717,7 +719,7 @@ void tst_qqmlcomponent::relativeUrl_data()
 {
     QTest::addColumn<QUrl>("url");
 
-#if !defined(Q_OS_ANDROID)
+#if !defined(Q_OS_ANDROID) && !defined(BUILTIN_TESTDATA)
     QTest::addRow("fromLocalFile") << QUrl::fromLocalFile("data/QtObjectComponent.qml");
     QTest::addRow("fromLocalFileHash") << QUrl::fromLocalFile("data/QtObjectComponent#2.qml");
     QTest::addRow("constructor") << QUrl("data/QtObjectComponent.qml");
@@ -1130,8 +1132,12 @@ void tst_qqmlcomponent::testSetInitialProperties()
 void tst_qqmlcomponent::createInsideJSModule()
 {
     QQmlEngine engine;
+    QString prefix;
+#if defined(Q_OS_ANDROID) || defined(BUILTIN_TESTDATA)
+    prefix = "qrc:/";
+#endif
     QQmlComponent component(&engine, testFileUrl("jsmodule/test.qml"));
-    QScopedPointer<QObject> root(component.create());
+    QScopedPointer<QObject> root(component.createWithInitialProperties({{"prefix", prefix}}));
     QVERIFY2(root, qPrintable(component.errorString()));
     QVERIFY(root->property("ok").toBool());
 }
@@ -1360,11 +1366,30 @@ void tst_qqmlcomponent::loadFromModule()
              name);
 }
 
+void tst_qqmlcomponent::loadFromModuleSyncAndAsync_data()
+{
+    loadFromModule_data();
+}
+
+void tst_qqmlcomponent::loadFromModuleSyncAndAsync()
+{
+    QFETCH(QString, uri);
+    QFETCH(QString, typeName);
+
+    QQmlEngine engine;
+    QQmlComponent syncAndAsync(&engine);
+    syncAndAsync.loadFromModule(uri, typeName, QQmlComponent::Asynchronous);
+    QVERIFY(syncAndAsync.isLoading());
+    syncAndAsync.loadFromModule(uri, typeName, QQmlComponent::PreferSynchronous);
+    QVERIFY2(syncAndAsync.isReady(), qPrintable(syncAndAsync.errorString()));
+    std::unique_ptr<QObject> object(syncAndAsync.create());
+    QVERIFY(object);
+}
+
 void tst_qqmlcomponent::loadFromModuleLifecycle()
 {
     QQmlEngine engine;
-    QList<int> loadFromModuleOrder;
-    QList<int> plainLoadOrder;
+    const QString text = "text"_L1;
     const QList<int> expected {1, 2, 3};
     {
         QQmlComponent component(&engine);
@@ -1373,19 +1398,58 @@ void tst_qqmlcomponent::loadFromModuleLifecycle()
         std::unique_ptr<QObject> root{ component.create() };
         LifeCycleWatcher *watcher = qobject_cast<LifeCycleWatcher *>(root.get());
         QVERIFY(watcher);
-        loadFromModuleOrder = watcher->states;
-        QCOMPARE(loadFromModuleOrder, expected);
+        QCOMPARE(watcher->states, expected);
+        QCOMPARE(watcher->observedTexts, QStringList(3));
+
+        const QString loaded = "load from module"_L1;
+        root.reset(component.createWithInitialProperties(QVariantMap{{text, loaded}}));
+        watcher = qobject_cast<LifeCycleWatcher *>(root.get());
+        QVERIFY(watcher);
+        QCOMPARE(watcher->states, expected);
+        QCOMPARE(watcher->observedTexts, QStringList({QString(), loaded, loaded}));
     }
+
     {
         QQmlComponent component(&engine);
         component.setData("import test; LifeCycleWatcher {}", {});
         QVERIFY2(component.isReady(), qPrintable(component.errorString()));
+
         std::unique_ptr<QObject> root{ component.create() };
         LifeCycleWatcher *watcher = qobject_cast<LifeCycleWatcher *>(root.get());
         QVERIFY(watcher);
-        plainLoadOrder = watcher->states;
+        QCOMPARE(watcher->states, expected);
+        QCOMPARE(watcher->observedTexts, QStringList(3));
+
+        const QString loaded = "load from data"_L1;
+        root.reset(component.createWithInitialProperties(QVariantMap{{text, loaded}}));
+        watcher = qobject_cast<LifeCycleWatcher *>(root.get());
+        QVERIFY(watcher);
+        QCOMPARE(watcher->states, expected);
+        QCOMPARE(watcher->observedTexts, QStringList({QString(), loaded, loaded}));
     }
-    QCOMPARE(loadFromModuleOrder, plainLoadOrder);
+
+    {
+        QQmlComponent component(&engine);
+        const QString compiled = "inline"_L1;
+        component.setData("import test; LifeCycleWatcher { text: 'inline' }", {});
+        QVERIFY2(component.isReady(), qPrintable(component.errorString()));
+
+        std::unique_ptr<QObject> root{ component.create() };
+        LifeCycleWatcher *watcher = qobject_cast<LifeCycleWatcher *>(root.get());
+        QVERIFY(watcher);
+        QCOMPARE(watcher->states, expected);
+        QCOMPARE(watcher->observedTexts, QStringList({QString(), compiled, compiled}));
+
+        const QString loaded = "overridden"_L1;
+        std::unique_ptr<QObject> withProperties(
+                component.createWithInitialProperties(QVariantMap{{text, loaded}}));
+        watcher = qobject_cast<LifeCycleWatcher *>(withProperties.get());
+        QVERIFY(watcher);
+        QCOMPARE(watcher->states, expected);
+        QCOMPARE(watcher->observedTexts, QStringList({QString(), loaded, loaded}));
+    }
+
+
 }
 
 struct CallVerifyingIncubtor : QQmlIncubator
