@@ -1,5 +1,5 @@
 // Copyright (C) 2016 The Qt Company Ltd.
-// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
 
 #include <QtTest/QtTest>
 #include <QtTest/QSignalSpy>
@@ -43,6 +43,8 @@ private slots:
     void cancel();
     void transformedpinchHandler_data();
     void transformedpinchHandler();
+    void dragVsPinch_data();
+    void dragVsPinch();
 
 private:
     QScopedPointer<QPointingDevice> touchscreen = QScopedPointer<QPointingDevice>(QTest::createTouchDevice());
@@ -215,14 +217,17 @@ void tst_QQuickPinchHandler::scale_data()
 {
     QTest::addColumn<QUrl>("qmlfile");
     QTest::addColumn<bool>("hasTarget");
-    QTest::newRow("targetModifying") << testFileUrl("pinchproperties.qml") << true;
-    QTest::newRow("nullTarget") << testFileUrl("nullTarget.qml") << false;
+    QTest::addColumn<bool>("axisEnabled");
+    QTest::newRow("targetModifying") << testFileUrl("pinchproperties.qml") << true << true;
+    QTest::newRow("nullTarget") << testFileUrl("nullTarget.qml") << false << true;
+    QTest::newRow("axisDiabled") << testFileUrl("pinchproperties.qml") << true << false;
 }
 
 void tst_QQuickPinchHandler::scale()
 {
     QFETCH(QUrl, qmlfile);
     QFETCH(bool, hasTarget);
+    QFETCH(bool, axisEnabled);
 
     QQuickView window;
     QVERIFY(QQuickTest::showView(window, qmlfile));
@@ -230,9 +235,10 @@ void tst_QQuickPinchHandler::scale()
     QVERIFY(root != nullptr);
     auto *pinchHandler = static_cast<PinchHandler *>(root->findChild<QQuickPinchHandler*>());
     QVERIFY(pinchHandler != nullptr);
+    pinchHandler->scaleAxis()->setEnabled(axisEnabled);
     QQuickItem *blackRect = (hasTarget ? pinchHandler->target() : pinchHandler->parentItem());
     QVERIFY(blackRect != nullptr);
-    QSignalSpy grabChangedSpy(pinchHandler, SIGNAL(grabChanged(QPointingDevice::GrabTransition, QEventPoint)));
+    QSignalSpy grabChangedSpy(pinchHandler, SIGNAL(grabChanged(QPointingDevice::GrabTransition,QEventPoint)));
     QSignalSpy scaleChangedSpy(pinchHandler, &QQuickPinchHandler::scaleChanged);
     if (lcPointerTests().isDebugEnabled()) QTest::qWait(500);
 
@@ -269,10 +275,13 @@ void tst_QQuickPinchHandler::scale()
     QCOMPARE(grabChangedSpy.size(), 3);
     QLineF line(p0, p1);
     const qreal startLength = line.length();
+    // to be redefined below
+    qreal lastScale = pinchHandler->persistentScale();
+    qreal expectedIncrement = 0;
 
     // move the same point even further and observe the change in scale
     for (int i = 0; i < 2; ++i) {
-        qreal lastScale = pinchHandler->activeScale();
+        lastScale = pinchHandler->activeScale();
         p1 += pd;
         pinchSequence.stationary(0).move(1, p1, &window).commit();
         QQuickTouchUtils::flush(&window);
@@ -282,21 +291,27 @@ void tst_QQuickPinchHandler::scale()
         qCDebug(lcPointerTests) << "pinchScale" << root->property("pinchScale").toReal()
                                 << "expected" << expectedScale << "; target scale" << blackRect->scale()
                                 << "increments" << scaleChangedSpy.size()
-                                << "multiplier" << scaleChangedSpy.last().first().toReal();
-        QVERIFY(qFloatDistance(root->property("pinchScale").toReal(), expectedScale) < 10);
-        QVERIFY(qFloatDistance(blackRect->scale(), expectedScale) < 10);
+                                << "multiplier" << (scaleChangedSpy.isEmpty() ? 1 : scaleChangedSpy.last().first().toReal());
+        if (axisEnabled) {
+            QVERIFY(qFloatDistance(root->property("pinchScale").toReal(), expectedScale) < 10);
+            QVERIFY(qFloatDistance(blackRect->scale(), expectedScale) < 10);
+        } else {
+            QCOMPARE(root->property("pinchScale").toInt(), 1);
+            QCOMPARE(blackRect->scale(), 1);
+        }
         QCOMPARE(pinchHandler->persistentScale(), root->property("pinchScale").toReal());
         QCOMPARE(pinchHandler->persistentScale(), pinchHandler->activeScale()); // in sync for the first gesture
         QCOMPARE(pinchHandler->scaleAxis()->persistentValue(), pinchHandler->activeScale());
         QCOMPARE(pinchHandler->scaleAxis()->activeValue(), pinchHandler->activeScale());
-        const qreal expectedIncrement = pinchHandler->activeScale() / lastScale;
-        QCOMPARE(scaleChangedSpy.size(), i + 1);
-        QCOMPARE(scaleChangedSpy.last().first().toReal(), expectedIncrement);
+        expectedIncrement = pinchHandler->activeScale() / lastScale;
+        QCOMPARE(scaleChangedSpy.size(), axisEnabled ? i + 1 : 0);
+        if (axisEnabled)
+            QCOMPARE(scaleChangedSpy.last().first().toReal(), expectedIncrement);
         QPointF expectedCentroid = p0 + (p1 - p0) / 2;
         QCOMPARE(pinchHandler->centroid().scenePosition(), expectedCentroid);
     }
 
-    qreal lastScale = pinchHandler->persistentScale();
+    lastScale = pinchHandler->persistentScale();
     pinchSequence.release(0, p0, &window).release(1, p1, &window).commit();
     QQuickTouchUtils::flush(&window);
     if (lcPointerTests().isDebugEnabled()) QTest::qWait(500);
@@ -326,21 +341,29 @@ void tst_QQuickPinchHandler::scale()
         pinchSequence.stationary(0).move(1, p1, &window).commit();
         QQuickTouchUtils::flush(&window);
         if (lcPointerTests().isDebugEnabled()) QTest::qWait(500);
-        QCOMPARE_GT(pinchHandler->persistentScale(), lastScale);
+        if (axisEnabled)
+            QCOMPARE_GT(pinchHandler->persistentScale(), lastScale);
+        else
+            QCOMPARE(pinchHandler->persistentScale(), 1);
         line.setP2(p1);
         qreal expectedActiveScale = line.length() / startLength;
         qCDebug(lcPointerTests) << i << "activeScale" << pinchHandler->activeScale()
                                 << "expected" << expectedActiveScale << "; scale" << pinchHandler->persistentScale()
                                 << "increments" << scaleChangedSpy.size()
-                                << "multiplier" << scaleChangedSpy.last().first().toReal();
-        QVERIFY(qFloatDistance(pinchHandler->activeScale(), expectedActiveScale) < 10);
+                                << "multiplier" << (scaleChangedSpy.isEmpty() ? 1 : scaleChangedSpy.last().first().toReal());
+        if (axisEnabled) {
+            QVERIFY(qFloatDistance(pinchHandler->activeScale(), expectedActiveScale) < 10);
+            QCOMPARE_NE(pinchHandler->persistentScale(), pinchHandler->activeScale()); // not in sync anymore
+        } else {
+            QCOMPARE(pinchHandler->persistentScale(), pinchHandler->activeScale());
+        }
         QCOMPARE(pinchHandler->persistentScale(), root->property("pinchScale").toReal());
         QCOMPARE(pinchHandler->scaleAxis()->persistentValue(), root->property("pinchScale").toReal());
-        QCOMPARE_NE(pinchHandler->persistentScale(), pinchHandler->activeScale()); // not in sync anymore
         QCOMPARE(pinchHandler->scaleAxis()->activeValue(), pinchHandler->activeScale());
-        const qreal expectedIncrement = pinchHandler->persistentScale() / lastScale;
-        QCOMPARE(scaleChangedSpy.size(), i + 3);
-        QCOMPARE(scaleChangedSpy.last().first().toReal(), expectedIncrement);
+        expectedIncrement = pinchHandler->persistentScale() / lastScale;
+        QCOMPARE(scaleChangedSpy.size(), axisEnabled ? i + 3 : 0);
+        if (axisEnabled)
+            QCOMPARE(scaleChangedSpy.last().first().toReal(), expectedIncrement);
     }
 
     // scale beyond maximumScale
@@ -349,12 +372,13 @@ void tst_QQuickPinchHandler::scale()
     pinchSequence.stationary(0).move(1, p1, &window).commit();
     QQuickTouchUtils::flush(&window);
     if (lcPointerTests().isDebugEnabled()) QTest::qWait(500);
-    QCOMPARE(blackRect->scale(), qreal(4));
-    QCOMPARE(pinchHandler->persistentScale(), qreal(4)); // limited by maximumScale
-    QCOMPARE(pinchHandler->scaleAxis()->persistentValue(), 4);
-    const qreal expectedIncrement = pinchHandler->activeScale() / lastScale;
-    QCOMPARE(scaleChangedSpy.size(), 5);
-    QCOMPARE(scaleChangedSpy.last().first().toReal(), expectedIncrement);
+    QCOMPARE(blackRect->scale(), axisEnabled ? 4 : 1);
+    QCOMPARE(pinchHandler->persistentScale(), axisEnabled ? 4 : 1); // limited by maximumScale
+    QCOMPARE(pinchHandler->scaleAxis()->persistentValue(), axisEnabled ? 4 : 1);
+    expectedIncrement = pinchHandler->activeScale() / lastScale;
+    QCOMPARE(scaleChangedSpy.size(), axisEnabled ? 5 : 0);
+    if (axisEnabled)
+        QCOMPARE(scaleChangedSpy.last().first().toReal(), expectedIncrement);
     pinchSequence.release(0, p0, &window).release(1, p1, &window).commit();
     QQuickTouchUtils::flush(&window);
     QCOMPARE(pinchHandler->active(), false);
@@ -557,19 +581,25 @@ void tst_QQuickPinchHandler::cumulativeNativeGestures_data()
     QTest::addColumn<const QPointingDevice*>("device");
     QTest::addColumn<Qt::NativeGestureType>("gesture");
     QTest::addColumn<qreal>("value");
+    QTest::addColumn<bool>("scalingEnabled");
+    QTest::addColumn<bool>("rotationEnabled");
     QTest::addColumn<QList<QPoint>>("expectedTargetTranslations");
 
     const auto *touchpadDevice = touchpad.get();
     const auto *mouse = QPointingDevice::primaryPointingDevice();
 
-    QTest::newRow("touchpad: rotate") << touchpadDevice << Qt::RotateNativeGesture << 5.0
+    QTest::newRow("touchpad: rotate") << touchpadDevice << Qt::RotateNativeGesture << 5.0 << true << true
                                       << QList<QPoint>{{-2, 2}, {-5, 4}, {-7, 6}, {-10, 7}};
-    QTest::newRow("touchpad: scale") << touchpadDevice << Qt::ZoomNativeGesture << 0.1
+    QTest::newRow("touchpad: scale") << touchpadDevice << Qt::ZoomNativeGesture << 0.1 << true << true
+                                     << QList<QPoint>{{3, 3}, {5, 5}, {8, 8}, {12, 12}};
+    QTest::newRow("touchpad: rotate disabled") << touchpadDevice << Qt::RotateNativeGesture << 5.0 << true << false
+                                      << QList<QPoint>{{-2, 2}, {-5, 4}, {-7, 6}, {-10, 7}};
+    QTest::newRow("touchpad: scale disabled") << touchpadDevice << Qt::ZoomNativeGesture << 0.1 << false << true
                                      << QList<QPoint>{{3, 3}, {5, 5}, {8, 8}, {12, 12}};
     if (mouse->type() == QInputDevice::DeviceType::Mouse) {
-        QTest::newRow("mouse: rotate") << mouse << Qt::RotateNativeGesture << 5.0
+        QTest::newRow("mouse: rotate") << mouse << Qt::RotateNativeGesture << 5.0 << true << true
                                        << QList<QPoint>{{-2, 2}, {-5, 4}, {-7, 6}, {-10, 7}};
-        QTest::newRow("mouse: scale") << mouse << Qt::ZoomNativeGesture << 0.1
+        QTest::newRow("mouse: scale") << mouse << Qt::ZoomNativeGesture << 0.1 << true << true
                                       << QList<QPoint>{{3, 3}, {5, 5}, {8, 8}, {12, 12}};
     } else {
         qCWarning(lcPointerTests) << "skipping mouse tests: primary device is not a mouse" << mouse;
@@ -581,6 +611,8 @@ void tst_QQuickPinchHandler::cumulativeNativeGestures()
     QFETCH(const QPointingDevice*, device);
     QFETCH(Qt::NativeGestureType, gesture);
     QFETCH(qreal, value);
+    QFETCH(bool, scalingEnabled);
+    QFETCH(bool, rotationEnabled);
     QFETCH(QList<QPoint>, expectedTargetTranslations);
 
     QCOMPARE(expectedTargetTranslations.size(), 4);
@@ -594,6 +626,8 @@ void tst_QQuickPinchHandler::cumulativeNativeGestures()
     QVERIFY(root != nullptr);
     QQuickPinchHandler *pinchHandler = root->findChild<QQuickPinchHandler*>("pinchHandler");
     QVERIFY(pinchHandler != nullptr);
+    pinchHandler->scaleAxis()->setEnabled(scalingEnabled);
+    pinchHandler->rotationAxis()->setEnabled(rotationEnabled);
     QQuickItem *target = root->findChild<QQuickItem*>("blackrect");
     QVERIFY(target != nullptr);
     QCOMPARE(pinchHandler->target(), target);
@@ -620,6 +654,10 @@ void tst_QQuickPinchHandler::cumulativeNativeGestures()
         default:
             break; // PinchHandler doesn't react to the others
         }
+        if (!rotationEnabled)
+            expectedRotation = 0;
+        if (!scalingEnabled)
+            expectedScale = 1;
 
         qCDebug(lcPointerTests) << i << gesture << "with value" << value
                                 << ": scale" << target->scale() << "expected" << expectedScale
@@ -641,8 +679,10 @@ void tst_QQuickPinchHandler::cumulativeNativeGestures()
         qCDebug(lcPointerTests) << "target moved by" << delta << "to" << target->position()
                                 << "active trans" << pinchHandler->activeTranslation()
                                 << "perst trans" << pinchHandler->persistentTranslation();
-        QCOMPARE_NE(target->position(), initialTargetPos);
-        QCOMPARE(delta.toPoint(), expectedTargetTranslations.at(i - 1));
+        if (scalingEnabled && rotationEnabled) {
+            QCOMPARE_NE(target->position(), initialTargetPos);
+            QCOMPARE(delta.toPoint(), expectedTargetTranslations.at(i - 1));
+        }
         // The native pinch gesture cannot include a translation component (and
         // the cursor doesn't move while you are performing the gesture on a touchpad).
         QCOMPARE(pinchHandler->activeTranslation(), QPointF());
@@ -1080,6 +1120,95 @@ void tst_QQuickPinchHandler::transformedpinchHandler()
     }
 }
 
+void tst_QQuickPinchHandler::dragVsPinch_data()
+{
+    // ptId is QEventPoint::id and also a 1-based index:
+    // 1, 2, 3 activate DragHandlers; 4, 5 are for the PinchHandler.
+    QTest::addColumn<int>("ptId1");
+    QTest::addColumn<int>("ptId2");
+    QTest::addColumn<QQuickPointerHandler::GrabPermission>("pinchGrabPermission");
+    QTest::addColumn<int>("expectedPinchActivations");
+
+    QTest::newRow("top two DragHandlers")
+            << 1 << 2 << QQuickPointerHandler::TakeOverForbidden << 0;
+    QTest::newRow("different DragHandlers")
+            << 2 << 3 << QQuickPointerHandler::TakeOverForbidden << 0;
+    QTest::newRow("one on DH, one on PH, TakeOverForbidden")
+            << 3 << 4 << QQuickPointerHandler::TakeOverForbidden << 0;
+    QTest::newRow("one on DH, one on PH, CanTakeOverFromAnything")
+            << 3 << 4 << QQuickPointerHandler::CanTakeOverFromAnything << 2;
+    QTest::newRow("both on PH")
+            << 4 << 5 << QQuickPointerHandler::TakeOverForbidden << 2;
+}
+
+void tst_QQuickPinchHandler::dragVsPinch()
+{
+    QFETCH(int, ptId1);
+    QFETCH(int, ptId2);
+    QFETCH(QQuickPointerHandler::GrabPermission, pinchGrabPermission);
+    QFETCH(int, expectedPinchActivations);
+
+    QQuickView window;
+    QVERIFY(QQuickTest::showView(window, testFileUrl("pinchAndDragHandlers.qml")));
+    QQuickItem *root = qobject_cast<QQuickItem*>(window.rootObject());
+    QVERIFY(root);
+    QQuickPinchHandler *pinchHandler = root->findChild<QQuickPinchHandler*>();
+    QVERIFY(pinchHandler);
+    pinchHandler->setGrabPermissions(pinchGrabPermission);
+    QSignalSpy pinchActiveSpy(pinchHandler, &QQuickPinchHandler::activeChanged);
+    QQuickMultiPointHandler *dh1 = root->findChild<QQuickMultiPointHandler*>("dh1");
+    QVERIFY(dh1);
+    QQuickMultiPointHandler *dh2 = root->findChild<QQuickMultiPointHandler*>("dh2");
+    QVERIFY(dh2);
+    QQuickMultiPointHandler *dh3 = root->findChild<QQuickMultiPointHandler*>("dh3");
+    QVERIFY(dh3);
+    const int dragThreshold = QGuiApplication::styleHints()->startDragDistance();
+    const QPoint rect1pos = dh1->parentItem()->position().toPoint();
+    const QPoint rect2pos = dh2->parentItem()->position().toPoint();
+    const QPoint rect3pos = dh3->parentItem()->position().toPoint();
+    const QPoint off = {10, 10}; // how far to press inside
+    const QList<QPoint> rectPos = {rect1pos, rect2pos, rect3pos};
+    const QList<QPoint> pointPos = {rect1pos + off, rect2pos + off, rect3pos + off,
+                                     rect2pos + QPoint(0, dh2->parentItem()->height() + 10),
+                                     rect3pos + QPoint(0, dh3->parentItem()->height() + 10)};
+    const QList<QQuickMultiPointHandler *> handlers = {dh1, dh2, dh3, pinchHandler, pinchHandler};
+
+    // press two points, one in each DragHandler's parent Rectangle
+    QPoint p1 = pointPos[ptId1 - 1];
+    QPoint p2 = pointPos[ptId2 - 1];
+    QTest::QTouchEventSequence pinchSequence = QTest::touchEvent(&window, touchscreen.get());
+    pinchSequence.press(ptId1, p1, &window).press(ptId2, p2, &window).commit();
+    QQuickTouchUtils::flush(&window);
+
+    qCDebug(lcPointerTests) << "press pts" << p1 << p2;
+    // drag outwards horizontally
+    for (int i = 1; i <= 4; ++i) {
+        p1 -= QPoint(dragThreshold, 0);
+        p2 += QPoint(dragThreshold, 0);
+        if (lcPointerTests().isDebugEnabled()) QTest::qWait(500);
+        pinchSequence.move(ptId1, p1, &window).move(ptId2, p2, &window).commit();
+        QQuickTouchUtils::flush(&window);
+        qCDebug(lcPointerTests) << i << "active" << dh1->active() << dh2->active() << dh3->active() << pinchHandler->active() << "pts" << p1 << p2
+                                << "rects @" << dh1->parentItem()->position() << dh2->parentItem()->position() << dh3->parentItem()->position();
+        if (i > 1 && !expectedPinchActivations) {
+            // We don't expect the PinchHandler to be active.  Check which DragHandlers are active.
+            if (ptId1 <= 3) {
+                QVERIFY(handlers[ptId1 - 1]->active());
+                QCOMPARE(handlers[ptId1 - 1]->parentItem()->position().x(), rectPos[ptId1 - 1].x() - dragThreshold * i);
+            }
+            if (ptId2 <= 3) {
+                QVERIFY(handlers[ptId2 - 1]->active());
+                QCOMPARE(handlers[ptId2 - 1]->parentItem()->position().x(), rectPos[ptId2 - 1].x() + dragThreshold * i);
+            }
+        }
+    }
+    if (lcPointerTests().isDebugEnabled()) QTest::qWait(500);
+    pinchSequence.release(ptId1, p1, &window).release(ptId2, p2, &window).commit();
+    // whether PinchHandler is activated depends on pinchHandler.grabPermissions
+    // and whether DragHandlers handle either or both points
+    QCOMPARE(pinchActiveSpy.size(), expectedPinchActivations);
+}
+
 QTEST_MAIN(tst_QQuickPinchHandler)
 
 #include "tst_qquickpinchhandler.moc"
