@@ -1,30 +1,5 @@
-﻿/****************************************************************************
-**
-** Copyright (C) 2019 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the test suite of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:GPL-EXCEPT$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3 as published by the Free Software
-** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2019 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
 
 #include <QtTest/QtTest>
 #include <QtQuick/qquickview.h>
@@ -87,6 +62,7 @@ private slots:
     void multilineDataTypes_data();
     void multilineDataTypes();
     void multilineStronglyTyped();
+    void fillTransform();
 
 private:
     QVector<QPolygonF> m_lowPolyLogo;
@@ -229,7 +205,7 @@ void tst_QQuickShape::changeSignals()
     QSignalSpy asyncPropSpy(obj, SIGNAL(asynchronousChanged()));
     obj->setAsynchronous(true);
     obj->setAsynchronous(false);
-    QCOMPARE(asyncPropSpy.count(), 2);
+    QCOMPARE(asyncPropSpy.size(), 2);
 
     QQmlListReference list(obj, "data");
     QQuickShapePath *vp = qobject_cast<QQuickShapePath *>(list.at(0));
@@ -249,47 +225,47 @@ void tst_QQuickShape::changeSignals()
     vp->setCapStyle(QQuickShapePath::RoundCap);
     vp->setDashOffset(10);
     vp->setDashPattern(QVector<qreal>() << 1 << 2 << 3 << 4);
-    QCOMPARE(strokeColorPropSpy.count(), 1);
-    QCOMPARE(vpChangeSpy.count(), 10);
+    QCOMPARE(strokeColorPropSpy.size(), 1);
+    QCOMPARE(vpChangeSpy.size(), 10);
 
     // Verify that property changes from Path and its elements bubble up and result in shapePathChanged().
     QQuickPath *path = vp;
     path->setStartX(30);
-    QCOMPARE(vpChangeSpy.count(), 11);
+    QCOMPARE(vpChangeSpy.size(), 11);
     QQmlListReference pathList(path, "pathElements");
     qobject_cast<QQuickPathLine *>(pathList.at(1))->setY(200);
-    QCOMPARE(vpChangeSpy.count(), 12);
+    QCOMPARE(vpChangeSpy.size(), 12);
 
     // Verify that property changes from the gradient bubble up and result in shapePathChanged().
     vp->setFillGradient(g);
-    QCOMPARE(vpChangeSpy.count(), 13);
+    QCOMPARE(vpChangeSpy.size(), 13);
     QQuickShapeLinearGradient *lgrad = qobject_cast<QQuickShapeLinearGradient *>(g);
     lgrad->setX2(200);
-    QCOMPARE(vpChangeSpy.count(), 14);
+    QCOMPARE(vpChangeSpy.size(), 14);
     QQmlListReference stopList(lgrad, "stops");
     QCOMPARE(stopList.count(), 5);
     qobject_cast<QQuickGradientStop *>(stopList.at(1))->setPosition(0.3);
-    QCOMPARE(vpChangeSpy.count(), 15);
+    QCOMPARE(vpChangeSpy.size(), 15);
     qobject_cast<QQuickGradientStop *>(stopList.at(1))->setColor(Qt::black);
-    QCOMPARE(vpChangeSpy.count(), 16);
+    QCOMPARE(vpChangeSpy.size(), 16);
+    vp->setFillTransform(QMatrix4x4(QTransform::fromScale(3, 0.14)));
+    QCOMPARE(vpChangeSpy.size(), 17);
 }
 
 void tst_QQuickShape::render()
 {
+    SKIP_IF_NO_WINDOW_GRAB;
+
     QScopedPointer<QQuickView> window(createView());
 
     window->setSource(testFileUrl("pathitem3.qml"));
     window->show();
     QVERIFY(QTest::qWaitForWindowExposed(window.data()));
 
-    if ((QGuiApplication::platformName() == QLatin1String("offscreen"))
-        || (QGuiApplication::platformName() == QLatin1String("minimal")))
-        QEXPECT_FAIL("", "Failure due to grabWindow not functional on offscreen/minimal platforms", Abort);
-
     QImage img = window->grabWindow();
     QVERIFY(!img.isNull());
 
-    QImage refImg(testFileUrl("pathitem3.png").toLocalFile());
+    QImage refImg(testFile("pathitem3.png"));
     QVERIFY(!refImg.isNull());
 
     QString errorMessage;
@@ -300,20 +276,18 @@ void tst_QQuickShape::render()
 
 void tst_QQuickShape::renderWithMultipleSp()
 {
+    SKIP_IF_NO_WINDOW_GRAB;
+
     QScopedPointer<QQuickView> window(createView());
 
     window->setSource(testFileUrl("pathitem4.qml"));
     window->show();
     QVERIFY(QTest::qWaitForWindowExposed(window.data()));
 
-    if ((QGuiApplication::platformName() == QLatin1String("offscreen"))
-        || (QGuiApplication::platformName() == QLatin1String("minimal")))
-        QEXPECT_FAIL("", "Failure due to grabWindow not functional on offscreen/minimal platforms", Abort);
-
     QImage img = window->grabWindow();
     QVERIFY(!img.isNull());
 
-    QImage refImg(testFileUrl("pathitem4.png").toLocalFile());
+    QImage refImg(testFile("pathitem4.png"));
     QVERIFY(!refImg.isNull());
 
     QString errorMessage;
@@ -324,20 +298,18 @@ void tst_QQuickShape::renderWithMultipleSp()
 
 void tst_QQuickShape::radialGrad()
 {
+    SKIP_IF_NO_WINDOW_GRAB;
+
     QScopedPointer<QQuickView> window(createView());
 
     window->setSource(testFileUrl("pathitem5.qml"));
     window->show();
     QVERIFY(QTest::qWaitForWindowExposed(window.data()));
 
-    if ((QGuiApplication::platformName() == QLatin1String("offscreen"))
-        || (QGuiApplication::platformName() == QLatin1String("minimal")))
-        QEXPECT_FAIL("", "Failure due to grabWindow not functional on offscreen/minimal platforms", Abort);
-
     QImage img = window->grabWindow();
     QVERIFY(!img.isNull());
 
-    QImage refImg(testFileUrl("pathitem5.png").toLocalFile());
+    QImage refImg(testFile("pathitem5.png"));
     QVERIFY(!refImg.isNull());
 
     QString errorMessage;
@@ -348,20 +320,18 @@ void tst_QQuickShape::radialGrad()
 
 void tst_QQuickShape::conicalGrad()
 {
+    SKIP_IF_NO_WINDOW_GRAB;
+
     QScopedPointer<QQuickView> window(createView());
 
     window->setSource(testFileUrl("pathitem6.qml"));
     window->show();
     QVERIFY(QTest::qWaitForWindowExposed(window.data()));
 
-    if ((QGuiApplication::platformName() == QLatin1String("offscreen"))
-        || (QGuiApplication::platformName() == QLatin1String("minimal")))
-        QEXPECT_FAIL("", "Failure due to grabWindow not functional on offscreen/minimal platforms", Abort);
-
     QImage img = window->grabWindow();
     QVERIFY(!img.isNull());
 
-    QImage refImg(testFileUrl("pathitem6.png").toLocalFile());
+    QImage refImg(testFile("pathitem6.png"));
     QVERIFY(!refImg.isNull());
 
     QString errorMessage;
@@ -372,20 +342,18 @@ void tst_QQuickShape::conicalGrad()
 
 void tst_QQuickShape::renderPolyline()
 {
+    SKIP_IF_NO_WINDOW_GRAB;
+
     QScopedPointer<QQuickView> window(createView());
 
     window->setSource(testFileUrl("pathitem7.qml"));
     window->show();
     QVERIFY(QTest::qWaitForWindowExposed(window.data()));
 
-    if ((QGuiApplication::platformName() == QLatin1String("offscreen"))
-        || (QGuiApplication::platformName() == QLatin1String("minimal")))
-        QEXPECT_FAIL("", "Failure due to grabWindow not functional on offscreen/minimal platforms", Abort);
-
     QImage img = window->grabWindow();
     QVERIFY(!img.isNull());
 
-    QImage refImg(testFileUrl("pathitem3.png").toLocalFile()); // It's a recreation of pathitem3 using PathPolyline
+    QImage refImg(testFile("pathitem3.png")); // It's a recreation of pathitem3 using PathPolyline
     QVERIFY(!refImg.isNull());
 
     QString errorMessage;
@@ -401,20 +369,18 @@ void tst_QQuickShape::renderPolyline()
 
 void tst_QQuickShape::renderMultiline()
 {
+    SKIP_IF_NO_WINDOW_GRAB;
+
     QScopedPointer<QQuickView> window(createView());
 
     window->setSource(testFileUrl("pathitem8.qml"));
     window->show();
     QVERIFY(QTest::qWaitForWindowExposed(window.data()));
 
-    if ((QGuiApplication::platformName() == QLatin1String("offscreen"))
-        || (QGuiApplication::platformName() == QLatin1String("minimal")))
-        QEXPECT_FAIL("", "Failure due to grabWindow not functional on offscreen/minimal platforms", Abort);
-
     QImage img = window->grabWindow();
     QVERIFY(!img.isNull());
 
-    QImage refImg(testFileUrl("pathitem8.png").toLocalFile());
+    QImage refImg(testFile("pathitem8.png"));
     QVERIFY(!refImg.isNull());
 
     QString errorMessage;
@@ -463,6 +429,8 @@ void tst_QQuickShape::polylineDataTypes_data()
 
 void tst_QQuickShape::polylineDataTypes()
 {
+    SKIP_IF_NO_WINDOW_GRAB;
+
     QFETCH(QVariant, path);
 
     QScopedPointer<QQuickView> window(createView());
@@ -473,14 +441,10 @@ void tst_QQuickShape::polylineDataTypes()
     window->show();
     QVERIFY(QTest::qWaitForWindowExposed(window.data()));
 
-    if ((QGuiApplication::platformName() == QLatin1String("offscreen"))
-        || (QGuiApplication::platformName() == QLatin1String("minimal")))
-        QEXPECT_FAIL("", "Failure due to grabWindow not functional on offscreen/minimal platforms", Abort);
-
     QImage img = window->grabWindow();
     QVERIFY(!img.isNull());
 
-    QImage refImg(testFileUrl("polyline.png").toLocalFile());
+    QImage refImg(testFile("polyline.png"));
     QVERIFY(!refImg.isNull());
 
     QString errorMessage;
@@ -614,6 +578,8 @@ void tst_QQuickShape::multilineDataTypes_data()
 
 void tst_QQuickShape::multilineDataTypes()
 {
+    SKIP_IF_NO_WINDOW_GRAB;
+
     QFETCH(QVariant, paths);
 
     QScopedPointer<QQuickView> window(createView());
@@ -624,14 +590,10 @@ void tst_QQuickShape::multilineDataTypes()
     window->show();
     QVERIFY(QTest::qWaitForWindowExposed(window.data()));
 
-    if ((QGuiApplication::platformName() == QLatin1String("offscreen"))
-        || (QGuiApplication::platformName() == QLatin1String("minimal")))
-        QEXPECT_FAIL("", "Failure due to grabWindow not functional on offscreen/minimal platforms", Abort);
-
     QImage img = window->grabWindow();
     QVERIFY(!img.isNull());
 
-    QImage refImg(testFileUrl("multiline.png").toLocalFile());
+    QImage refImg(testFile("multiline.png"));
     QVERIFY(!refImg.isNull());
 
     QString errorMessage;
@@ -662,6 +624,8 @@ void tst_QQuickShape::multilineDataTypes()
 
 void tst_QQuickShape::multilineStronglyTyped()
 {
+    SKIP_IF_NO_WINDOW_GRAB;
+
     QScopedPointer<QQuickView> window(createView());
     window->setSource(testFileUrl("multilineStronglyTyped.qml"));
     QQuickShape *shape = qobject_cast<QQuickShape *>(window->rootObject());
@@ -672,14 +636,10 @@ void tst_QQuickShape::multilineStronglyTyped()
     QVERIFY(QTest::qWaitForWindowExposed(window.data()));
     provider->setPaths(m_lowPolyLogo);
 
-    if ((QGuiApplication::platformName() == QLatin1String("offscreen"))
-            || (QGuiApplication::platformName() == QLatin1String("minimal")))
-        QEXPECT_FAIL("", "Failure due to grabWindow not functional on offscreen/minimal platforms", Abort);
-
     QImage img = window->grabWindow();
     QVERIFY(!img.isNull());
 
-    QImage refImg(testFileUrl("multiline.png").toLocalFile());
+    QImage refImg(testFile("multiline.png"));
     QVERIFY(!refImg.isNull());
 
     QString errorMessage;
@@ -708,6 +668,40 @@ void tst_QQuickShape::multilineStronglyTyped()
     }
 }
 
+void tst_QQuickShape::fillTransform()
+{
+    QScopedPointer<QQuickView> window(createView());
+
+    window->setSource(testFileUrl("filltransform.qml"));
+    qApp->processEvents();
+
+    QQuickShape *obj = findItem<QQuickShape>(window->rootObject(), "shape1");
+    QVERIFY(obj != nullptr);
+    QQmlListReference list(obj, "data");
+    QCOMPARE(list.count(), 2);
+
+    QQuickShapePath *p1 = qobject_cast<QQuickShapePath *>(list.at(0));
+    QVERIFY(p1 != nullptr);
+    QVERIFY(p1->objectName() == "path1");
+    QVERIFY(p1->fillTransform() == QMatrix4x4(2,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1));
+
+    QQuickShapePath *p2 = qobject_cast<QQuickShapePath *>(list.at(1));
+    QVERIFY(p2 != nullptr);
+    QVERIFY(p2->objectName() == "path2");
+    QVERIFY(p2->fillTransform().isIdentity());
+
+    QMatrix4x4 xf(QTransform::fromTranslate(-36, 0).shear(0.35, 0));
+    p1->setFillTransform(xf);
+    QVERIFY(p1->fillTransform() == xf);
+
+    QVERIFY(p2->fillTransform().isIdentity());
+    p2->setFillTransform(xf);
+    QVERIFY(p2->fillTransform() == xf);
+
+    p1->setFillTransform(QMatrix4x4{});
+    QVERIFY(p1->fillTransform().isIdentity());
+}
+
 QTEST_MAIN(tst_QQuickShape)
 
 #include "tst_qquickshape.moc"
